from .module import Module


class Papertrail(Module):
    dependencies = ['task']

    def __init__(self, **kwargs):
        super().__init__('papertrail', **kwargs)

    def add_arguments(self, parser):
        subp = parser.add_subparsers(help='papertrail help')

        p = subp.add_parser('add')
        p.set_defaults(papertrail_handler=self.handle_add)

        p = subp.add_parser('connect')
        p.add_argument('target')
        p.set_defaults(papertrail_handler=self.handle_connect)

    def handle_add(self, args):
        task_mod = self.client.get_module('task')
        task_mod.add('papertrail', '!gliderlabs/logspout:latest',
                     mode='global')
        task_mod.mount_set('/var/run/docker.sock', '/var/run/docker.sock',
                           type='bind')
        task_mod.command('syslog+tls://{}'.format(args.log))



import os
from configparser import ConfigParser

import boto3

from .module import Module
from .utils import entity_already_exists, limit_exceeded


class Profile(Module):
    def __init__(self, **kwargs):
        super().__init__('profile', **kwargs)
        self.aws_path = os.path.expanduser('~/.aws/credentials')

    def add_arguments(self, parser):
        subp = parser.add_subparsers(help='profile help')

        p = subp.add_parser('ls', help='list profiles')
        p.add_argument('name', metavar='NAME', nargs='?', help='profile name')
        p.set_defaults(profile_handler=self.handle_list)

        # p = subp.add_parser('mk', help='make a new profile')
        # p.add_argument('name', metavar='NAME', help='profile name')
        # p.set_defaults(profile_handler=self.handle_make)

        # p = subp.add_parser('rm', help='remove a profile')
        # p.add_argument('name', metavar='NAME', help='profile name')
        # p.set_defaults(profile_handler=self.handle_remove)

        p = subp.add_parser('bucket', help='set FUKU bucket')
        p.add_argument('name', metavar='NAME', help='bucket name')
        p.set_defaults(profile_handler=self.handle_bucket)

        p = subp.add_parser('sl', help='select a profile')
        p.add_argument('name', metavar='NAME', help='profile name')
        p.set_defaults(profile_handler=self.handle_select)

        p = subp.add_parser('sh', help='show selected profile')
        p.set_defaults(profile_handler=self.handle_show)

    def handle_list(self, args):
        self.list(args.name)

    def list(self, name):
        for prof in self.list_local_profiles():
            if not name or name == prof:
                print(prof)

    def handle_make(self, args):
        self.make(args.name)

    def make(self, name):
        self.use_context = False
        if name not in self.list_local_profiles():
            self.error(f'please create a "{name}" profile using AWS CLI then rerun')
        self.create_ec2_role(name)

    def handle_select(self, args):
        self.select(args.name)

    def select(self, name):
        if name and name not in self.list_local_profiles():
            self.error(f'no profile named "{name}"')
        self.store_set('selected', name)
        self.clear_parent_selections()

    def handle_show(self, args):
        self.show()

    def show(self):
        sel = self.store_get('selected')
        if sel:
            print(sel)

    def handle_bucket(self, args):
        self.bucket(args.name)

    def bucket(self, name):
        s3 = self.get_boto_resource('s3', {'profile': self.get_selected()})
        bucket = s3.Bucket(name)
        try:
            bucket.load()
        except:
            bucket.create()
        self.store_set('bucket', name)

    def list_local_profiles(self):
        cfg = ConfigParser()
        cfg.read(self.aws_path)
        return cfg.sections()

    def create_ec2_role(self, user):
        role_name = 'ec2-role'
        inst_name = 'ec2-profile'
        iam = boto3.Session(profile_name=user).client('iam')
        self.create_role(user, role_name, ['ec2-policy'], iam=iam)
        with entity_already_exists():
            iam.create_instance_profile(
                InstanceProfileName=inst_name
            )
        # TODO: Add AmazonEC2ContainerServiceforEC2Role policy to role
        with limit_exceeded():
            iam.add_role_to_instance_profile(
                InstanceProfileName=inst_name,
                RoleName=role_name
            )

    def create_role(self, user, name, policies=[], iam=None):
        if iam is None:
            iam = boto3.Session(profile_name=user).client('iam')
        with entity_already_exists():
            iam.create_role(
                RoleName=name,
                AssumeRolePolicyDocument=f'file://{self.data_path("ecs-assume-role.json")}'
            )
        for policy in policies:
            ctx = self.get_context({
                'name': name,
            })
            with self.template_file(f'{policy}.json', ctx) as policy_file:
                iam.put_role_policy(
                    RoleName=name,
                    PolicyName=policy,
                    PolicyDocument=policy_file
                )

    def delete_role(self, user, role):
        self.run(
            '$aws iam delete-role'
            ' --profile {user}'
            ' --role-name {role}'.format(
                user=user,
                role=role
            )
        )

    def delete_ec2_role(self, user):
        role_name = 'ec2-role'
        inst_name = 'ec2-profile'
        self.run(
            '$aws iam delete-instance-profile'
            ' --profile {user}'
            ' --instance-profile-name {inst_name}'.format(
                user=user,
                inst_name=inst_name
            )
        )
        self.delete_role(user, role_name)

    def get_user_id(self):
        profile = self.get_selected()
        sts = self.get_boto_client('sts')
        try:
            return sts.get_caller_identity()['Account']
        except:
            return None

    def get_my_context(self):
        ctx = {}
        sel = self.get_selected()
        if sel:
            ctx['profile'] = sel
        if 'bucket' not in self.store:
            self.error('bucket not set')
        ctx['bucket'] = self.store['bucket']
        return ctx

    def get_selected(self):
        sel = self.store_get('selected')
        if not sel:
            self.error('no profile currently selected')
        return sel



from .module import Module
# from .utils import dict_to_env, dict_to_ports


class SSL(Module):
    dependencies = ['task']

    def __init__(self, **kwargs):
        super().__init__('ssl', **kwargs)

    def add_arguments(self, parser):
        subp = parser.add_subparsers(help='ssl help')

        p = subp.add_parser('add')
        # p.add_argument('task')
        p.add_argument('email')
        p.add_argument('domain')
        p.add_argument('upstream')
        p.add_argument('--staging', '-s', action='store_true')
        p.add_argument('--update', '-u', action='store_true')
        p.set_defaults(ssl_handler=self.handle_add)

    def handle_add(self, args):
        task_mod = self.client.get_module('task')
        if not args.update:
            task_mod.add('ssl', '!smashwilson/lets-nginx')
        env = {
            'EMAIL': args.email,
            'DOMAIN': args.domain,
            'UPSTREAM': args.upstream  # prefix for substitution
        }
        if args.staging:
            env['STAGING'] = '1'
        else:
            task_mod.env_unset('ssl', ['STAGING'])
        task_mod.env_set('ssl', env)
        ports = {
            '80': '80',
            '443': '443'
        }
        task_mod.ports_set('ssl', ports)
        volumes = {
            'letsencrypt': {
                'container': '/etc/letsencrypt'
            },
            'letsencrypt-backups': {
                'container': '/var/lib/letsencrypt'
            },
            'dhparam-cache': {
                'container': '/cache'
            }
        }
        task_mod.volume_add('ssl', 'letsencrypt', '/etc/letsencrypt')
        task_mod.volume_add('ssl', 'letsencrypt-backups', '/var/lib/letsencrypt')
        task_mod.volume_add('ssl', 'dhparam-cache', '/cache')
        # ctr = args.upstream.split(':')[0]
        # task_mod.link('ssl', ctr)
        # TODO: Remove reverse link.



import re
import json
import base64

from .module import Module


class Node(Module):
    dependencies = ['cluster']
    ami_map = {
      # 'ap-southeast-2': 'ami-862211e5',  # docker enabled amazon
        'ap-southeast-2': 'ami-e7878484',  # arch linux
    }

    def __init__(self, **kwargs):
        super().__init__('node', **kwargs)

    def add_arguments(self, parser):
        subp = parser.add_subparsers(help='node help')

        p = subp.add_parser('ls', help='list nodes')
        p.add_argument('name', metavar='NAME', nargs='?', help='node name')
        p.set_defaults(node_handler=self.handle_list)

        p = subp.add_parser('mk', help='make a node')
        p.add_argument('name', metavar='NAME', help='node name')
        p.add_argument('--manager', '-m', action='store_true', help='manager node (Swarm)')
        p.add_argument('--availability-zone', '-a', choices=['a', 'b'], help='availability zone')
        p.set_defaults(node_handler=self.handle_make)

        p = subp.add_parser('rm', help='remove a node')
        p.add_argument('name', metavar='NAME', help='node name')
        p.set_defaults(node_handler=self.handle_remove)

        p = subp.add_parser('bastion', help='make a bastion node')
        p.set_defaults(node_handler=self.handle_bastion)

        p = subp.add_parser('ssh', help='SSH to a node')
        p.add_argument('name', metavar='NAME', nargs='?', help='node name')
        p.set_defaults(node_handler=self.handle_ssh)

        p = subp.add_parser('init', help='initialise swarm')
        p.add_argument('name', metavar='NAME', help='node name')
        p.set_defaults(node_handler=self.handle_init_swarm)

        p = subp.add_parser('join', help='join swarm')
        p.add_argument('name', metavar='NAME', help='node name')
        p.set_defaults(node_handler=self.handle_join_swarm)

        p = subp.add_parser('wait', help='wait for OK status')
        p.add_argument('name', metavar='NAME', help='node name')
        p.set_defaults(node_handler=self.handle_wait)

        p = subp.add_parser('reboot', help='reboot a node')
        p.add_argument('name', metavar='NAME', help='node name')
        p.set_defaults(node_handler=self.handle_reboot)

    def handle_list(self, args):
        self.list(args.name)

    def list(self, name):
        for name in self.get_instance_names():
            print(name)

    def handle_make(self, args):
        self.make(args.name, args.manager)

    def make(self, name, manager=False):
        self.validate(name)
        existing = self.get_instance_names()
        if name in existing:
            self.error(f'node "{name}" already exists')
        if not len(existing) and not manager:
            self.error(f'must have at least one manager before workers')
        ctx = self.get_context({'node': name})
        ec2 = self.get_boto_client('ec2')
        image = self.ami_map[ctx['region']]
        sg_id = self.get_module('cluster').get_security_group_id()
        sn = self.get_module('cluster').get_private_subnet()
        opts = {
            'ImageId': image,
            'SubnetId': sn.id,
            'KeyName': f'fuku-{ctx["cluster"]}',
            'SecurityGroupIds': [sg_id],
            'InstanceType': 't2.micro',
            'IamInstanceProfile': {
                'Name': 'ec2-profile'
            },
            'MinCount': 1,
            'MaxCount': 1
        }
        with self.template_file('arch-user-data.sh', ctx) as user_data:
            opts['UserData'] = user_data
            inst = ec2.run_instances(**opts)
        inst_id = inst['Instances'][0]['InstanceId']
        self.tag_instance(inst_id, name, manager, ec2=ec2, ctx=ctx)
        self.wait(name)
        if manager:
            self.init_swarm(name)

    def handle_remove(self, args):
        self.remove(args.name)

    def remove(self, name):
        inst = self.get_instance(name)
        inst.terminate()

    def handle_init_swarm(self, args):
        self.init_swarm(args.name)

    def init_swarm(self, name):
        ctx = self.get_context()
        inst = self.get_instance(name)
        ip = inst.private_ip_address
        resp = self.ssh_run(
            f'docker swarm init --advertise-addr {ip}',
            name,
            capture='text'
        )
        try:
            token = re.search(r'token\s+(.*)\s+\\', resp).group(1)
            port = re.search(r':(\d\d\d\d)', resp).group(1)
        except AttributeError:
            self.error('swarm initialisation failed')
        # s3 = self.get_boto_client('s3')
        # try:
        #     data = s3.get_object(
        #         Bucket=ctx['bucket'],
        #         Key=f'fuku/{ctx["cluster"]}/swarm.json',
        #     )['Body'].read().decode()
        #     data = json.loads(data)
        # except:
        #     data = {
        #         'managers': []
        #     }
        # data['managers'].append({
        #     'workertoken': token,
        #     'ip': ip,
        #     'port': port
        # })
        # s3.put_object(
        #     Bucket=ctx['bucket'],
        #     Key=f'fuku/{ctx["cluster"]}/swarm.json',
        #     Body=json.dumps(data)
        # )
        ec2 = self.get_boto_client('ec2')
        ec2.create_tags(
            Resources=[inst.id],
            Tags=[
                {'Key': 'swarmtoken', 'Value': token},
                {'Key': 'swarmport', 'Value': port}
            ]
        )
        self.ssh_run(
            'docker network create --driver overlay all',
            name=name,
            capture='discard'
        )

    def handle_join_swarm(self, args):
        self.join_swarm(args.name)

    def join_swarm(self, name):
        for inst in self.iter_managers():
            try:
                token, port = None, None
                for tag in inst.tags:
                    if tag['Key'] == 'swarmtoken':
                        token = tag['Value']
                    if tag['Key'] == 'swarmport':
                        port = tag['Value']
                if token is None:
                    raise Exception
                resp = self.ssh_run(
                    f'docker swarm join --token {token} {inst.private_ip_address}:{port}',
                    name=name,
                    capture='text'
                )
                if resp != 'This node joined a swarm as a worker.':
                    self.error('failed to join swarm')
            except:
                pass
            else:
                break

    def handle_wait(self, args):
        self.wait(args.name)

    def wait(self, name):
        inst = self.get_instance(name)
        ec2 = self.get_boto_client('ec2')
        waiter = ec2.get_waiter('instance_status_ok')
        waiter.wait(InstanceIds=[inst.id])

    def handle_ssh(self, args):
        self.ssh_run('', args.name, tty=True)

    def handle_reboot(self, args):
        self.reboot(args.name)

    def reboot(self, name):
        ctx = self.get_context()
        try:
            name = name or ctx['node']
        except KeyError:
            self.error('unknown node')
        inst = self.get_instance(name)
        inst.reboot()

    def tag_instance(self, inst_id, name, manager=False, ec2=None, ctx=None):
        if ctx is None:
            ctx = self.get_context()
        if ec2 is None:
            ec2 = self.get_boto_resource('ec2')
        ec2.create_tags(
            Resources=[inst_id],
            Tags=[
                {'Key': 'Name', 'Value': f'fuku-{ctx["cluster"]}-{name}'},
                {'Key': 'name', 'Value': name},
                {'Key': 'cluster', 'Value': ctx['cluster']},
                {'Key': 'node', 'Value': 'manager' if manager else 'worker'}
            ]
        )

    def iter_instances(self):
        ec2 = self.get_boto_resource('ec2')
        cluster = self.client.get_selected('cluster')
        filters = [
            {
                'Name': 'tag:cluster',
                'Values': [cluster]
            },
            {
                'Name': 'instance-state-name',
                'Values': ['pending', 'running', 'stopping', 'stopped']
            }
        ]
        for inst in ec2.instances.filter(Filters=filters):
            yield inst

    def get_instance_names(self):
        names = []
        for inst in self.iter_instances():
            for tag in inst.tags:
                if tag['Key'] == 'name':
                    names.append(tag['Value'])
                    break
        return names

    def get_instance(self, name):
        ec2 = self.get_boto_resource('ec2')
        cluster = self.client.get_selected('cluster')
        filters = [
            {
                'Name': 'tag:cluster',
                'Values': [cluster]
            },
            {
                'Name': 'tag:name',
                'Values': [name]
            },
            {
                'Name': 'instance-state-name',
                'Values': ['pending', 'running', 'shutting-down', 'stopping', 'stopped']
            }
        ]
        insts = list(ec2.instances.filter(Filters=filters))
        if not len(insts):
            self.error(f'no node in cluster "{cluster}" with name "{name}"')
        return insts[0]

    def get_bastion(self):
        return self.get_instance('bastion')

    def iter_managers(self):
        ctx = self.get_context()
        ec2 = self.get_boto_resource('ec2')
        filters = [
            {
                'Name': 'tag:cluster',
                'Values': [ctx['cluster']]
            },
            {
                'Name': 'tag:node',
                'Values': ['manager']
            },
            {
                'Name': 'instance-state-name',
                'Values': ['running']
            }
        ]
        for inst in ec2.instances.filter(Filters=filters):
            yield inst

    def iter_nodes(self):
        ctx = self.get_context()
        ec2 = self.get_boto_resource('ec2')
        filters = [
            {
                'Name': 'tag:cluster',
                'Values': [ctx['cluster']]
            },
            {
                'Name': 'instance-state-name',
                'Values': ['running']
            }
        ]
        for inst in ec2.instances.filter(Filters=filters):
            yield inst

    def ssh_run(self, cmd, name=None, inst=None, tty=False, capture=None):
        ctx = self.get_context()
        if inst is None:
            name = name or ctx['node']
            inst = self.get_instance(name)
        bastion = self.get_bastion()
        ip = bastion.public_ip_address
        priv_ip = inst.private_ip_address
        b64cmd = base64.b64encode(cmd.encode()).decode()
        full_cmd = f'ssh-add {ctx["pem"]} && ssh{" -t" if tty else ""} -o "StrictHostKeyChecking no" -A ec2-user@{ip} ssh{" -t" if tty else ""} -o \\\"StrictHostKeyChecking no\\\" {priv_ip} "\`echo {b64cmd} | base64 -di\`"'
        return self.run(full_cmd, capture=capture)

    def handle_bastion(self, args):
        self.bastion()

    def bastion(self):
        name = 'bastion'
        ctx = self.get_context()
        ec2 = self.get_boto_client('ec2')
        image = self.ami_map[ctx['region']]
        sg_id = self.get_module('cluster').get_security_group_id()
        sn = self.get_module('cluster').get_public_subnet()
        opts = {
            'ImageId': image,
            'NetworkInterfaces': [
                {
                    'DeviceIndex': 0,
                    'SubnetId': sn.id,
                    'Groups': [sg_id],
                    'AssociatePublicIpAddress': True
                }
            ],
            'KeyName': f'fuku-{ctx["cluster"]}',
            # 'SecurityGroupIds': [sg_id],
            'InstanceType': 't2.nano',
            'IamInstanceProfile': {
                'Name': 'ec2-profile'
            },
            'MinCount': 1,
            'MaxCount': 1
        }
        inst = ec2.run_instances(**opts)
        inst_id = inst['Instances'][0]['InstanceId']
        ec2.create_tags(
            Resources=[inst_id],
            Tags=[
                {'Key': 'Name', 'Value': f'fuku-{ctx["cluster"]}-{name}'},
                {'Key': 'cluster', 'Value': f'{ctx["cluster"]}'},
                {'Key': 'name', 'Value': f'{name}'},
                {'Key': 'bastion', 'Value': 'true'}
            ]
        )
        self.wait(name)
        return inst_id

    def mgr_run(self, cmd, tty=False, capture=None):
        try:
            mgr = list(self.iter_managers())[0]
        except IndexError:
            self.error('no managers available')
        return self.ssh_run(cmd, inst=mgr, tty=tty, capture=capture)

    def all_run(self, cmd):
        for node in self.iter_nodes():
            self.ssh_run(cmd, inst=node, capture=False)

    def get_my_context(self):
        ctx = {}
        node = self.store_get('selected')
        if node is not None:
            ctx['node'] = node
        return ctx


class EcsNode(Node):
    ami_map = {
        'ap-southeast-2': 'ami-fbe9eb98'
    }

    def add_arguments(self, parser):
        subp = parser.add_subparsers(help='node help')

        p = subp.add_parser('ls', help='list nodes')
        p.add_argument('name', metavar='NAME', nargs='?', help='node name')
        p.set_defaults(node_handler=self.handle_list)

        p = subp.add_parser('mk', help='make a node')
        p.add_argument('name', metavar='NAME', help='node name')
        p.add_argument('--availability-zone', '-a', choices=['a', 'b'], help='availability zone')
        p.set_defaults(node_handler=self.handle_make)

        p = subp.add_parser('rm', help='remove a node')
        p.add_argument('name', metavar='NAME', help='node name')
        p.set_defaults(node_handler=self.handle_remove)

        p = subp.add_parser('bastion', help='make a bastion node')
        p.set_defaults(node_handler=self.handle_bastion)

        p = subp.add_parser('ssh', help='SSH to a node')
        p.add_argument('name', metavar='NAME', nargs='?', help='node name')
        p.set_defaults(node_handler=self.handle_ssh)

        p = subp.add_parser('wait', help='wait for OK status')
        p.add_argument('name', metavar='NAME', help='node name')
        p.set_defaults(node_handler=self.handle_wait)

        p = subp.add_parser('reboot', help='reboot a node')
        p.add_argument('name', metavar='NAME', help='node name')
        p.set_defaults(node_handler=self.handle_reboot)

    def handle_make(self, args):
        self.make(args.name, args.availability_zone)

    def make(self, name, zone):
        self.validate(name)
        existing = self.get_instance_names()
        if name in existing:
            self.error(f'node "{name}" already exists')
        ctx = self.get_context({'node': name})
        ec2 = self.get_boto_client('ec2')
        image = self.ami_map[ctx['region']]
        sg_id = self.get_module('cluster').get_security_group_id()
        sn = self.get_module('cluster').get_private_subnet(zone=zone)
        opts = {
            'ImageId': image,
            'SubnetId': sn.id,
            'KeyName': f'fuku-{ctx["cluster"]}',
            'SecurityGroupIds': [sg_id],
            'InstanceType': 't2.micro',
            'IamInstanceProfile': {
                'Name': 'ec2-profile'
            },
            'MinCount': 1,
            'MaxCount': 1,
            'UserData': f'#!/bin/bash\necho ECS_CLUSTER=fuku-{ctx["cluster"]} >> /etc/ecs/ecs.config'
        }
        inst = ec2.run_instances(**opts)
        inst_id = inst['Instances'][0]['InstanceId']
        self.tag_instance(inst_id, name, False, ec2=ec2, ctx=ctx)
        self.wait(name)



import argparse
import string
import random
from contextlib import contextmanager

import botocore


class EntityAlreadyExists(Exception):
    pass


class StoreKeyValuePair(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None, sep='='):
        if not isinstance(values, list):
            values = [values]
        for pair in values:
            k, v = pair.split(sep)
        setattr(namespace, self.dest, dict([(k, v) for k, v in [p.split(sep) for p in values]]))


class StorePortPair(StoreKeyValuePair):
    def __call__(self, *args, **kwargs):
        super().__call__(*args, sep=':', **kwargs)


def env_to_dict(env):
    res = {}
    for pair in env:
        res[pair['name']] = pair['value']
    return res


def dict_to_env(val):
    env = []
    for k, v in val.items():
        env.append({'name': k, 'value': v})
    return env


def env_to_string(env, opt='-e'):
    if env:
        return ' ' + ' '.join([
            '%s %s="%s"' % (opt, k, v) for k, v in env_to_dict(env).items()
        ])
    else:
        return ''


def ports_to_dict(env):
    res = {}
    for pair in env:
        res[pair['hostPort']] = pair['containerPort']
    return res


def dict_to_ports(val):
    env = []
    for k, v in val.items():
        env.append({'hostPort': int(k), 'containerPort': int(v), 'protocol': 'tcp'})
    return env


def ports_to_string(val, opt='-p'):
    if val:
        return ' ' + ' '.join([
            '%s %s:%s' % (opt, k, v) for k, v in ports_to_dict(val).items()
        ])
    else:
        return ''


def volumes_to_dict(volumes):
    res = {}
    for info in volumes:
        res[info['name']] = info.get('host', {}).get('sourcePath', None)
    return res


def dict_to_volumes(val):
    volumes = []
    for k, v in val.items():
        item = {'name': k}
        if v:
            item['host'] = {'sourcePath': v}
        volumes.append(item)
    return volumes


def mounts_to_dict(mounts):
    res = {}
    for info in mounts:
        res[info['sourceVolume']] = {
            'containerPath': info['containerPath'],
            'readOnly': info['readOnly']
        }
    return res


def dict_to_mounts(val):
    mounts = []
    for k, v in val.items():
        mounts.append({
            'sourceVolume': k,
            'containerPath': v['containerPath'],
            'readOnly': v.get('readOnly', False)
        })
    return mounts


def mounts_to_string(val, existing={}, opt='--mount'):
    if val:
        mounts = mounts_to_dict(val)
        elems = []
        for src, dst in mounts.items():
            if src not in existing:
                elems.append('%s src=%s,dst=%s' % (opt, src, dst))
        for src, dst in existing.items():
            if src not in mounts.keys():
                elems.append('--mount-rm %s' % dst)
        return ' ' + ' '.join(elems)
    else:
        return ''


def gen_secret(length=64):
    return ''.join(random.SystemRandom().choice(string.ascii_uppercase + string.digits) for _ in range(length))


def gen_name(length=16):
    return ''.join(
        random.SystemRandom().choice(string.ascii_uppercase) + 
        random.SystemRandom().choice(string.ascii_uppercase + string.digits) for _ in range(length - 1)
    )


@contextmanager
def entity_already_exists(hide=True):
    try:
        yield
    except botocore.exceptions.ClientError as e:
        if e.response['Error']['Code'] not in [
                'EntityAlreadyExists',
                'InvalidKeyPair.Duplicate',
                'InvalidGroup.Duplicate',
                'InvalidPermission.Duplicate',
                'RepositoryAlreadyExistsException',
                'ResourceAlreadyExistsException'
        ]:
            raise
        if not hide:
            raise EntityAlreadyExists


@contextmanager
def limit_exceeded():
    try:
        yield
    except botocore.exceptions.ClientError as e:
        if e.response['Error']['Code'] != 'LimitExceeded':
            raise



import re
import os
import sys
from contextlib import contextmanager
import subprocess

from fabric.operations import (
    _prefix_env_vars, _prefix_commands,
    _AttributeString
)
from fabric.state import env as _env, output, win32


class CommandError(Exception):
    def __init__(self, out):
        super().__init__(out.stderr)
        self.out = out


def local(command, capture=False, shell=None, ignore_errors=False, env=None):
    """
    Run a command on the local system.
    `local` is simply a convenience wrapper around the use of the builtin
    Python ``subprocess`` module with ``shell=True`` activated. If you need to
    do anything special, consider using the ``subprocess`` module directly.
    ``shell`` is passed directly to `subprocess.Popen
    <http://docs.python.org/library/subprocess.html#subprocess.Popen>`_'s
    ``execute`` argument (which determines the local shell to use.)  As per the
    linked documentation, on Unix the default behavior is to use ``/bin/sh``,
    so this option is useful for setting that value to e.g.  ``/bin/bash``.
    `local` is not currently capable of simultaneously printing and
    capturing output, as `~fabric.operations.run`/`~fabric.operations.sudo`
    do. The ``capture`` kwarg allows you to switch between printing and
    capturing as necessary, and defaults to ``False``.
    When ``capture=False``, the local subprocess' stdout and stderr streams are
    hooked up directly to your terminal, though you may use the global
    :doc:`output controls </usage/output_controls>` ``output.stdout`` and
    ``output.stderr`` to hide one or both if desired. In this mode, the return
    value's stdout/stderr values are always empty.
    When ``capture=True``, you will not see any output from the subprocess in
    your terminal, but the return value will contain the captured
    stdout/stderr.
    In either case, as with `~fabric.operations.run` and
    `~fabric.operations.sudo`, this return value exhibits the ``return_code``,
    ``stderr``, ``failed``, ``succeeded``, ``command`` and ``real_command``
    attributes. See `run` for details.
    `~fabric.operations.local` will honor the `~fabric.context_managers.lcd`
    context manager, allowing you to control its current working directory
    independently of the remote end (which honors
    `~fabric.context_managers.cd`).
    .. versionchanged:: 1.0
        Added the ``succeeded`` and ``stderr`` attributes.
    .. versionchanged:: 1.0
        Now honors the `~fabric.context_managers.lcd` context manager.
    .. versionchanged:: 1.0
        Changed the default value of ``capture`` from ``True`` to ``False``.
    .. versionadded:: 1.9
        The return value attributes ``.command`` and ``.real_command``.
    """
    given_command = command
    # Apply cd(), path() etc
    with_env = _prefix_env_vars(command, local=True)
    wrapped_command = _prefix_commands(with_env, 'local')
    # if output.debug:
    #     print("[localhost] local: %s" % (wrapped_command))
    # elif output.running:
    #     print("[localhost] local: " + given_command)
    # Tie in to global output controls as best we can; our capture argument
    # takes precedence over the output settings.
    dev_null = None
    if capture:
        out_stream = subprocess.PIPE
        err_stream = subprocess.PIPE
    else:
        dev_null = open(os.devnull, 'w+')
        # Non-captured, hidden streams are discarded.
        out_stream = None if output.stdout else dev_null
        err_stream = None if output.stderr else dev_null
    if env is None:
        env = os.environ
    try:
        cmd_arg = wrapped_command if win32 else [wrapped_command]
        p = subprocess.Popen(cmd_arg, shell=True, stdout=out_stream,
                             stderr=err_stream, executable=shell,
                             close_fds=(not win32),
                             env=env)
        (stdout, stderr) = p.communicate()
    finally:
        if dev_null is not None:
            dev_null.close()
    # Handle error condition (deal with stdout being None, too)
    out = _AttributeString(stdout.decode().strip() if stdout else "")
    err = _AttributeString(stderr.decode().strip() if stderr else "")
    out.command = given_command
    out.real_command = wrapped_command
    out.failed = False
    out.return_code = p.returncode
    out.stderr = err
    if p.returncode not in _env.ok_ret_codes and not ignore_errors:
        out.failed = True
        # msg = "local() encountered an error (return code %s) while executing '%s'" % (p.returncode, command)
        # print('\n\n')
        # if out:
        #     print('{}'.format(out))
        # print('\n')
        # if err:
        #     print('{}'.format(err))
        # sys.exit(1)
        raise CommandError(out)
        # error(message=msg, stdout=out, stderr=err)
    out.succeeded = not out.failed
    # If we were capturing, this will be a string; otherwise it will be None.
    return out

run = local


@contextmanager
def already_exists(expr):
    try:
        yield
    except Exception as e:
        if not re.search(expr, str(e)):
            raise

# from fabric.api import task, run as fabrun


# @task
# def run(cmd):
#     fabrun(cmd)



#from contextlib import contextmanager

# # from fabric.operations import _run_command
# from fabric.context_managers import (
#     warn_only as warn_only_manager,
#     quiet as quiet_manager
# )


# @contextmanager
# def _noop():
#     yield


# env = {
#     'use_shell': False,
    
# }


# def _shell_wrap(command, shell_escape, shell=True, sudo_prefix=None):
#     """
#     Conditionally wrap given command in env.shell (while honoring sudo.)
#     """
#     # Honor env.shell, while allowing the 'shell' kwarg to override it (at
#     # least in terms of turning it off.)
#     if shell and not env.use_shell:
#         shell = False
#     # Sudo plus space, or empty string
#     if sudo_prefix is None:
#         sudo_prefix = ""
#     else:
#         sudo_prefix += " "
#     # If we're shell wrapping, prefix shell and space. Next, escape the command
#     # if requested, and then quote it. Otherwise, empty string.
#     if shell:
#         shell = env.shell + " "
#         if shell_escape:
#             command = _shell_escape(command)
#         command = '"%s"' % command
#     else:
#         shell = ""
#     # Resulting string should now have correct formatting
#     return sudo_prefix + shell + command


# def _run_command(command, shell=True, pty=True, combine_stderr=True,
#                  sudo=False, user=None, quiet=False, warn_only=False,
#                  stdout=None, stderr=None, group=None, timeout=None,
#                  shell_escape=None, capture_buffer_size=None):
#     """
#     Underpinnings of `run` and `sudo`. See their docstrings for more info.
#     """
#     manager = _noop
#     if warn_only:
#         manager = warn_only_manager
#     # Quiet's behavior is a superset of warn_only's, so it wins.
#     if quiet:
#         manager = quiet_manager
#     with manager():
#         # Set up new var so original argument can be displayed verbatim later.
#         given_command = command

#         # # Check if shell_escape has been overridden in env
#         # if shell_escape is None:
#         #     shell_escape = env.get('shell_escape', True)

#         # Handle context manager modifications, and shell wrapping
#         wrapped_command = _shell_wrap(
#             _prefix_env_vars(_prefix_commands(command, 'remote')),
#             shell_escape,
#             shell,
#             _sudo_prefix(user, group) if sudo else None
#         )
#         # Execute info line
#         which = 'sudo' if sudo else 'run'
#         if output.debug:
#             print("[%s] %s: %s" % (env.host_string, which, wrapped_command))
#         elif output.running:
#             print("[%s] %s: %s" % (env.host_string, which, given_command))

#         # Actual execution, stdin/stdout/stderr handling, and termination
#         result_stdout, result_stderr, status = _execute(
#             channel=default_channel(), command=wrapped_command, pty=pty,
#             combine_stderr=combine_stderr, invoke_shell=False, stdout=stdout,
#             stderr=stderr, timeout=timeout,
#             capture_buffer_size=capture_buffer_size)

#         # Assemble output string
#         out = _AttributeString(result_stdout)
#         err = _AttributeString(result_stderr)

#         # Error handling
#         out.failed = False
#         out.command = given_command
#         out.real_command = wrapped_command
#         if status not in env.ok_ret_codes:
#             out.failed = True
#             msg = "%s() received nonzero return code %s while executing" % (
#                 which, status
#             )
#             if env.warn_only:
#                 msg += " '%s'!" % given_command
#             else:
#                 msg += "!\n\nRequested: %s\nExecuted: %s" % (
#                     given_command, wrapped_command
#                 )
#             error(message=msg, stdout=out, stderr=err)

#         # Attach return code to output string so users who have set things to
#         # warn only, can inspect the error code.
#         out.return_code = status

#         # Convenience mirror of .failed
#         out.succeeded = not out.failed

#         # Attach stderr for anyone interested in that.
#         out.stderr = err

#         return out

# run = _run_command



from .module import Module
from .runner import CommandError
from .utils import (
    env_to_string, ports_to_string, env_to_dict, dict_to_env,
    mounts_to_string, volumes_to_dict
)


class Service(Module):
    dependencies = ['task']

    def __init__(self, **kwargs):
        super().__init__('service', **kwargs)

    def add_arguments(self, parser):
        subp = parser.add_subparsers(help='service help')

        p = subp.add_parser('ls', help='list services')
        p.add_argument('task', metavar='TASK', nargs='?', help='task name')
        p.set_defaults(service_handler=self.handle_list)

        p = subp.add_parser('mk', help='make a service')
        p.add_argument('task', metavar='TASK', help='task name')
        p.add_argument('--replicas', '-r', help='number of replicas')
        p.set_defaults(service_handler=self.handle_make)

        p = subp.add_parser('up', help='update a service')
        p.add_argument('task', metavar='TASK', help='task name')
        p.add_argument('--replicas', '-r', help='number of replicas')
        p.set_defaults(service_handler=self.handle_update)

        p = subp.add_parser('scale', help='scale a service')
        p.add_argument('task', metavar='TASK', help='task name')
        p.add_argument('replicas', metavar='REPLICAS', help='number of replicas')
        p.set_defaults(service_handler=self.handle_scale)

        p = subp.add_parser('redeploy', help='redeploy all')
        p.set_defaults(service_handler=self.handle_redeploy)

        p = subp.add_parser('rm', help='remove a service')
        p.add_argument('task', metavar='TASK', help='task name')
        p.set_defaults(service_handler=self.handle_remove)

        p = subp.add_parser('run', help='run a command')
        p.add_argument('task', metavar='TASK', help='task name')
        p.add_argument('command', metavar='COMMAND', help='command to run')
        p.set_defaults(service_handler=self.handle_run)

        # p = subp.add_parser('remove')
        # p.add_argument('task')
        # p.add_argument('--volumes', '-v', action='store_true', help='remove volumes')
        # p.set_defaults(service_handler=self.handle_remove)

        # p = subp.add_parser('logs', help='get logs')
        # p.add_argument('name', help='container name')
        # p.set_defaults(service_handler=self.handle_logs)

    def handle_list(self, args):
        self.list(args.task)

    def list(self, task_name):
        ctx = self.get_context()
        node_mod = self.client.get_module('node')
        if task_name:
            svc_name = self.get_name()
            try:
                node_mod.mgr_run(
                    f'docker service inspect {svc_name}',
                    capture=False
                )
            except CommandError:
                pass
        else:
            node_mod.mgr_run(
                'docker service ls',
                capture=False
            )

    def handle_make(self, args):
        self.make(args.task, args.replicas)

    def make(self, task_name, replicas=None):
        ctx = self.get_context()
        task_mod = self.client.get_module('task')
        app_task = task_mod.get_app_task()
        try:
            env = env_to_dict(task_mod.get_container_definition(app_task, '_', fail=False)['environment'])
        except TypeError:
            env = {}
        ctr_def = task_mod.get_container_definition(app_task, task_name)
        svc_name = self.get_name()
        cmd = f'$(aws --region {ctx["region"]} ecr get-login);'
        cmd += f' docker pull {ctr_def["image"]};'
        cmd += f' docker service create'
        cmd += f' --name {svc_name}'
        if replicas:
            cmd += f' --replicas {replicas}'
        if ctr_def.get('cpu', None):
            cmd += f' --reserve-cpu {ctr_def["cpu"]}'
        if ctr_def.get('memoryReservation', None):
            cmd += f' --reserve-memory {int(ctr_def["memoryReservation"]) * 1048576}'  # convert to bytes
        cmd += ' --with-registry-auth'
        env.update(env_to_dict(ctr_def.get('environment', [])))
        cmd += env_to_string(
            dict_to_env(env),
            opt='-e'
        )
        cmd += ports_to_string(
            ctr_def.get('portMappings', []),
            opt='-p'
        )
        cmd += ' --network all'
        cmd += mounts_to_string(
            ctr_def.get('mountPoints', {}),
            self.get_mounts(task_name),
            opt='--mount'
        )
        cmd += ' ' + ctr_def['image']
        if ctr_def.get('command', None):
            cmd += ' ' + ' '.join(ctr_def['command'])
        cmd = '\'' + cmd + '\''
        self.make_volumes(volumes_to_dict(app_task.get('volumes', [])))
        node_mod = self.client.get_module('node')
        try:
            node_mod.mgr_run(
                cmd,
                capture='discard'
            )
        except CommandError as e:
            self.error('failed to launch service, please check the task command')

    def handle_update(self, args):
        self.update(args.task, args.replicas)

    def update(self, task_name, replicas=None):
        ctx = self.get_context()
        task_mod = self.client.get_module('task')
        app_task = task_mod.get_app_task()
        try:
            env = env_to_dict(task_mod.get_container_definition(app_task, '_', fail=False)['environment'])
        except TypeError:
            env = {}
        ctr_def = task_mod.get_container_definition(app_task, task_name)
        svc_name = self.get_name()
        cmd = f'$(aws --region {ctx["region"]} ecr get-login);'
        cmd += f' docker pull {ctr_def["image"]};'
        cmd += f' docker service update'
        if replicas:
            cmd += f' --replicas {replicas}'
        if ctr_def.get('cpu', None):
            cmd += f' --reserve-cpu {ctr_def["cpu"]}'
        if ctr_def.get('memoryReservation', None):
            cmd += f' --reserve-memory {int(ctr_def["memoryReservation"]) * 1048576}'  # convert to bytes
        cmd += ' --with-registry-auth'
        env.update(env_to_dict(ctr_def.get('environment', [])))
        cmd += env_to_string(
            dict_to_env(env),
            opt='--env-add'
        )
        cmd += ports_to_string(
            ctr_def.get('portMappings', []),
            opt='--publish-add'
        )
        cmd += mounts_to_string(
            ctr_def.get('mountPoints', {}),
            self.get_mounts(task_name),
            opt='--mount-add'
        )
        cmd += ' --force'
        cmd += ' --image ' + ctr_def['image']
        if ctr_def.get('command', None):
            cmd += ' --args "' + ' '.join(ctr_def['command']) + '"'
        cmd += f' {svc_name}'
        cmd = '\'' + cmd + '\''
        self.make_volumes(volumes_to_dict(app_task.get('volumes', [])))
        node_mod = self.client.get_module('node')
        try:
            node_mod.mgr_run(
                cmd,
                capture='discard'
            )
        except CommandError as e:
            self.error('failed to update service, please check the task command')

    def handle_remove(self, args):
        self.remove(args.task)

    # def handle_remove(self, args):
    #     mach_mod = self.get_module('node')
    #     mach = mach_mod.get_selected()
    #     mach_mod.ssh_run(
    #         'docker service rm {}'.format(args.task),
    #         name=mach,
    #         capture='discard'
    #     )
    #     if args.volumes:
    #         task_mod = self.client.get_module('task')
    #         task = task_mod.get_task(task_mod.get_task_name())
    #         self.delete_volumes(volumes_to_dict(task.get('volumes', [])))

    def handle_run(self, args):
        self.run(args.task, args.command)

    def run(self, task_name, command):
        task_mod = self.get_module('task')
        ecs_cli = self.get_boto_client('ecs')
        ec2 = self.get_boto_resource('ec2')
        ctx = self.get_context()
        family = '_' + task_mod.get_task_family(task_name)
        cluster = f'fuku-{ctx["cluster"]}'
        task_arns = ecs_cli.list_tasks(
            cluster=cluster,
            family=family,
            desiredStatus='RUNNING'
        )['taskArns']
        if not task_arns:
            self.error('no running tasks for that service')
        # task_id = task_arns[0][task_arns[0].rfind('/') + 1:]
        cinst_arn = ecs_cli.describe_tasks(
            cluster=cluster,
            tasks=[
                task_arns[0]
            ]
        )['tasks'][0]['containerInstanceArn']
        inst_id = ecs_cli.describe_container_instances(
            cluster=cluster,
            containerInstances=[
                cinst_arn
            ]
        )['containerInstances'][0]['ec2InstanceId']
        inst = ec2.Instance(inst_id)
        cmd = f'docker exec -it `docker ps | grep {family} | awk \'{{ print $1 }}\'` {command}'
        node_mod = self.get_module('node')
        node_mod.ssh_run(cmd, inst=inst, tty=True)

    def handle_logs(self, args):
        mach_mod = self.client.get_module('machine')
        mach = mach_mod.get_selected()
        cmd = 'docker logs {}'.format(args.name)
        mach_mod.ssh_run(
            cmd,
            name=mach,
            tty=True,
            capture=False
        )

    def handle_redeploy(self, args):
        self.redeploy()

    def redeploy(self):
        for svc in self.iter_services():
            self.update(svc)

    def get_task(self, name, escape=True):
        task_mod = self.client.get_module('task')
        return task_mod.get_task(name, escape)

    def get_container_definition(self, task, name):
        task_mod = self.client.get_module('task')
        return task_mod.get_container_definition(task, name)

    def make_volumes(self, volumes):
        node_mod = self.client.get_module('node')
        for name, src in volumes.items():
            cmd = f'docker volume create --name {name}'
            try:
                mach_mod.mgr_run(
                    cmd,
                    capture='discard'
                )
            except self.CommandError:
                pass

    def delete_volumes(self, volumes):
        mach_mod = self.client.get_module('machine')
        mach = mach_mod.get_selected()
        for name, src in volumes.items():
            cmd = 'docker volume rm {}'.format(name)
            try:
                mach_mod.ssh_run(
                    cmd,
                    name=mach,
                    capture='discard'
                )
            except self.CommandError:
                pass

    def get_mounts(self, task):
        node_mod = self.client.get_module('node')
        try:
            data = node_mod.mgr_run(
                f'docker service inspect {task}',
                capture='json'
            )
        except CommandError:
            data = None
        if data:
            mounts = data[0]['Spec']['TaskTemplate']['ContainerSpec'].get('Mounts', [])
            mounts = dict([(m['Source'], m['Target']) for m in mounts])
            return mounts
        else:
            return {}

    def get_name(self, task_name):
        ctx = self.get_context()
        return f'{ctx["app"]}-{task_name}'

    def dependency_removal(self, module, id):
        if module.name == 'task':
            if self.is_running(id):
                return [
                    {
                        'module': self,
                        'id': id
                    }
                ]
        return []


class EcsService(Service):
    def add_arguments(self, parser):
        subp = parser.add_subparsers(help='service help')

        p = subp.add_parser('ls', help='list services')
        p.add_argument('task', metavar='TASK', nargs='?', help='task name')
        p.set_defaults(service_handler=self.handle_list)

        p = subp.add_parser('mk', help='make a service')
        p.add_argument('task', metavar='TASK', help='task name')
        p.add_argument('--expose', '-e', action='store_true', help='expose through load-balancer')
        p.add_argument('--replicas', '-r', help='number of replicas')
        p.set_defaults(service_handler=self.handle_make)

        p = subp.add_parser('up', help='update a service')
        p.add_argument('task', metavar='TASK', help='task name')
        p.add_argument('--replicas', '-r', help='number of replicas')
        p.set_defaults(service_handler=self.handle_update)

        p = subp.add_parser('scale', help='scale a service')
        p.add_argument('task', metavar='TASK', help='task name')
        p.add_argument('replicas', metavar='REPLICAS', help='number of replicas')
        p.set_defaults(service_handler=self.handle_scale)

        p = subp.add_parser('redeploy', help='redeploy all')
        p.set_defaults(service_handler=self.handle_redeploy)

        p = subp.add_parser('rm', help='remove a service')
        p.add_argument('task', metavar='TASK', help='task name')
        p.set_defaults(service_handler=self.handle_remove)

        p = subp.add_parser('run', help='run a command')
        p.add_argument('task', metavar='TASK', help='task name')
        p.add_argument('command', metavar='COMMAND', help='command to run')
        p.set_defaults(service_handler=self.handle_run)

    def list(self, task_name):
        for svc in self.iter_services(task_name):
            print(svc)

    def handle_make(self, args):
        self.make(args.task, args.replicas, args.expose)

    def make(self, task_name, replicas=None, expose=False):
        ctx = self.get_context()
        cluster = f'fuku-{ctx["cluster"]}'
        task_mod = self.client.get_module('task')
        app_task = task_mod.get_task(None)
        task = task_mod.get_task(task_name)
        task['family'] = '_' + task['family']
        env = env_to_dict(task_mod.get_container_definition(app_task, ctx['app'])['environment'])
        ctr_def = task_mod.get_container_definition(task, task_name)
        env.update(env_to_dict(ctr_def['environment']))
        ctr_def['environment'] = dict_to_env(env)
        task['containerDefinitions'] = [ctr_def]
        ecs_cli = self.get_boto_client('ecs')
        skip = set(['taskDefinitionArn', 'revision', 'status', 'requiresAttributes'])
        task = ecs_cli.register_task_definition(**dict([
            (k, v) for k, v in task.items() if k not in skip
        ]))['taskDefinition']
        # TODO: Deregister previous task definitions.
        kwargs = {
            'cluster': cluster,
            'serviceName': f'fuku-{ctx["app"]}-{task_name}',
            'taskDefinition': f'{task["family"]}:{task["revision"]}',
            'desiredCount': int(replicas) if replicas is not None else 1,
            'deploymentConfiguration': {
                'maximumPercent': 200,
                'minimumHealthyPercent': 50
            },
            'placementStrategy': [{
                'type': 'spread',
                'field': 'attribute:ecs.availability-zone'
            }]
        }
        if expose:
            kwargs['loadBalancers'] = [
                {
                    'targetGroupArn': self.get_module('app').get_target_group_arn(),
                    'containerName': task_name,
                    'containerPort': 80
                }
            ]
            kwargs['role'] = 'ecsServiceRole'
        ecs_cli.create_service(**kwargs)

    def update(self, task_name, replicas=None):
        ctx = self.get_context()
        cluster = f'fuku-{ctx["cluster"]}'
        task_mod = self.client.get_module('task')
        app_task = task_mod.get_task(None)
        task = task_mod.get_task(task_name)
        task['family'] = '_' + task['family']
        env = env_to_dict(task_mod.get_container_definition(app_task, ctx['app'])['environment'])
        ctr_def = task_mod.get_container_definition(task, task_name)
        env.update(env_to_dict(ctr_def['environment']))
        ctr_def['environment'] = dict_to_env(env)
        task['containerDefinitions'] = [ctr_def]
        ecs_cli = self.get_boto_client('ecs')
        skip = set(['taskDefinitionArn', 'revision', 'status', 'requiresAttributes'])
        task = ecs_cli.register_task_definition(**dict([
            (k, v) for k, v in task.items() if k not in skip
        ]))['taskDefinition']
        # TODO: Deregister previous task definitions.
        kwargs = {
            'cluster': cluster,
            'service': f'fuku-{ctx["app"]}-{task_name}',
            'taskDefinition': f'{task["family"]}:{task["revision"]}',
        }
        if replicas:
            kwargs['desiredCount'] = int(replicas) if replicas is not None else 1
        ecs_cli.update_service(**kwargs)

    def handle_scale(self, args):
        self.scale(args.task, args.replicas)

    def scale(self, task_name, replicas=None):
        ctx = self.get_context()
        cluster = f'fuku-{ctx["cluster"]}'
        family = f'_fuku-{ctx["cluster"]}-{ctx["app"]}-{task_name}'
        ecs_cli = self.get_boto_client('ecs')
        task = ecs_cli.describe_task_definition(
            taskDefinition=family
        )['taskDefinition']
        ecs_cli.update_service(
            cluster=cluster,
            service=f'fuku-{ctx["app"]}-{task_name}',
            taskDefinition=f'{family}:{task["revision"]}',
            desiredCount=int(replicas) if replicas is not None else 1
        )

    def remove(self, task_name):
        self.confirm_remove(task_name)
        self.scale(task_name, 0)
        ctx = self.get_context()
        cluster = f'fuku-{ctx["cluster"]}'
        ecs_cli = self.get_boto_client('ecs')
        svc = f'fuku-{ctx["app"]}-{task_name}'
        # waiter = ecs_cli.get_waiter('services_inactive')
        # waiter.wait(
        #     cluster=cluster,
        #     services=[svc]
        # )
        ecs_cli.delete_service(
            cluster=cluster,
            service=svc
        )

    def iter_services(self, task_name=None):
        ecs_cli = self.get_boto_client('ecs')
        paginator = ecs_cli.get_paginator('list_services')
        ctx = self.get_context()
        cluster = f'fuku-{ctx["cluster"]}'
        services = paginator.paginate(cluster=cluster)
        for svcs in services:
            for s in svcs['serviceArns']:
                ii = s.rfind('-')
                yield s[ii + 1:]

    def is_running(self, task_name):
        ecs_cli = self.get_boto_client('ecs')
        ctx = self.get_context()
        results = ecs_cli.describe_services(
            cluster=f'fuku-{ctx["cluster"]}',
            services=[
                f'fuku-{ctx["app"]}-{task_name}'
            ]
        )
        try:
            return len(results['services']) > 0
        except KeyError:
            return False



from .module import Module


class Metrics(Module):
    dependencies = ['service']
    metric_choices = [
        'memory',
        'cpu'
    ]

    def __init__(self, **kwargs):
        super().__init__('metrics', **kwargs)

    def add_arguments(self, parser):
        subp = parser.add_subparsers(help='metrics help')

        p = subp.add_parser('ls', help='list metrics')
        p.add_argument('--all', '-a', action='store_true', help='include inactive metrics')
        p.set_defaults(metrics_handler=self.handle_list)

        p = subp.add_parser('mk', help='add metrics')
        p.add_argument('task', metavar='TASK', help='task name')
        p.add_argument('metric', metavar='METRIC', choices=self.metric_choices, help='metric')
        p.set_defaults(metrics_handler=self.handle_make)

        p = subp.add_parser('rm')
        p.add_argument('name', metavar='NAME', help='metric to remove')
        p.set_defaults(metrics_handler=self.handle_remove)

        p = subp.add_parser('clear')
        p.set_defaults(metrics_handler=self.handle_clear)

    def handle_list(self, args):
        self.list(args.all)

    def list(self, all):
        ctx = self.get_context()
        metrics = self.gets3(f'{ctx["cluster"]}/metrics.json') or {}
        for task, mets in metrics.items():
            print(f'{task} ({",".join( sorted( mets ) )})')

    def handle_make(self, args):
        self.make(args.task, args.metric)

    def make(self, task_name, met_name):
        ctx = self.get_context()
        metrics = self.gets3(f'{ctx["cluster"]}/metrics.json') or {}
        met_list = metrics.setdefault(task_name, [])
        if met_name in met_list:
            return
        met_list.append(met_name)
        self.puts3(f'{ctx["cluster"]}/metrics.json', metrics)
        met = f'docker-{ctx["app"]}-{task_name}.*-{met_name}.percent-'
        node_mod = self.client.get_module('node')
        cmd = f'echo {met} >> /usr/share/collectd/collectd-cloudwatch/src/cloudwatch/config/whitelist.conf'
        cmd += '; systemctl restart collectd'
        cmd = '\'' + cmd + '\''
        node_mod.all_run(cmd)

    def handle_remove(self, args):
        mach_mod = self.client.get_module('machine')
        cmd = 'echo args.name >> /usr/share/collectd/collectd-cloudwatch/src/cloudwatch/config/whitelist.conf'
        mach_mod.ssh_run(cmd, discard=False)

    def handle_clear(self, args):
        self.clear()

    def clear(self):
        ctx = self.get_context()
        self.puts3(f'{ctx["cluster"]}/metrics.json', {})
        cmd = f'truncate -s 0 /usr/share/collectd/collectd-cloudwatch/src/cloudwatch/config/whitelist.conf'
        cmd += '; systemctl restart collectd'
        cmd = '\'' + cmd + '\''
        node_mod = self.client.get_module('node')
        node_mod.all_run(cmd)



import os, sys
from setuptools import setup, find_packages

version = '0.0.11'

setup(
    name='fuku',
    version=version,
    author='Luke Hodkinson',
    author_email='furious.luke@gmail.com',
    maintainer='Luke Hodkinson',
    maintainer_email='furious.luke@gmail.com',
    url='https://github.com/ABASystems/python-fuku',
    description='Easier Docker based deployments to AWS.',
    long_description=open(os.path.join(os.path.dirname(__file__), 'README.md')).read(),
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5'
    ],
    license='BSD',
    packages=find_packages(),
    include_package_data=True,
    package_data={'': ['*.txt', '*.js', '*.html', '*.*']},
    install_requires=[
        'setuptools',
        'six',
        'boto3',
        'fabric3'
    ],
    scripts=[
        'fuku/scripts/fuku',
    ],
    zip_safe=False,
)
